import mlflow
import argparse
from aml_auth import start_mlflow_experiment, end_mlflow_experiment, get_ml_client
from assessment_flow import test_benchmark
from aml_fetch_data import fetch_data
from evaluation import eval_7class, eval_binary


def exp_benchmark(ml_client, dataset_name, min_score, max_score, sample_size, model, prompt_template, eval_func, n_prompts_per_call, seed=1234):
    df = fetch_data(ml_client, dataset_name)
    acc = test_benchmark(f"./outputs/{dataset_name}", sample_size, model, prompt_template, min_score, max_score, df, eval_func, n_prompts_per_call, seed)
    # log in mlflow
    mlflow.log_metric(f"{dataset_name}-acc", acc)
    return acc


def exp_all_groundedness_benchmarks(ml_client, sample_size, model, prompt_template):
    # save template in outputs for record, and log it in mlflow
    f = open("./outputs/prompt_template.txt", "a")
    f.write(prompt_template)
    f.close()
    mlflow.log_param("template", prompt_template[:500])

    openai_acc = exp_benchmark(ml_client, "groundedness-openai", 1, 7, sample_size, model, prompt_template, eval_7class, 1)
    fever_acc = exp_benchmark(ml_client, "groundedness-fever", 0, 1, sample_size, model, prompt_template, eval_binary, 5)
    docnli_acc = exp_benchmark(ml_client, "groundedness-docnli", 0, 1, sample_size, model, prompt_template, eval_binary, 5)
    mctest_acc = exp_benchmark(ml_client, "groundedness-mctest", 0, 1, sample_size, model, prompt_template, eval_binary, 5)
    mnli_acc = exp_benchmark(ml_client, "groundedness-mnli", 0, 1, sample_size, model, prompt_template, eval_binary, 5)

    # get average
    groundedness_avg_acc = (openai_acc+fever_acc+docnli_acc+mctest_acc+mnli_acc)/5
    mlflow.log_metric("groundedness-avg-acc", groundedness_avg_acc)
    return groundedness_avg_acc


# command line args
# argv[1] area (groundedness / contentharm/ ...)
# argv[2] number of samples per benchmark
# argv[3] name of experiment
# argv[4] name of txt containing prompt
# argv[5] name of model

# we do not enable repeated sampling and CI in local mode
# because the expected sample size in this mode is not large and CI is likely to be large and uninformative
# for now, not enabling selective datasets - always run on all datasets
# for now, not enabling user-input batch size bc this can have negative effect on results
# for now, fixing boilerplate for batching

# if experiment for any benchmark crashes, skip and keep running others. All results are periodically saved in csv
# upon failure direct user to the csv, but still emit metrics of the good benchmarks to aml

# locally will create an experiment folder with 2 subdirectories: datasets and results. datasets will be downloaded from
# adls for experiments and results serve as a checkpoint, storing every sample's true label and gpt label incrementally
# inside results, each experiment has a folder with name exp_<guid>, so that it will not be overwritten by later experiments

def main():
    # Instantiate the parser
    print("entered e2e exp")
    parser = argparse.ArgumentParser()
    parser.add_argument('--area', type=str, help='groundedness/contentharm/...')
    parser.add_argument('--samples_per_benchmark', type=int, help='number of samples per benchmark')
    parser.add_argument('--experiment_name', type=str, help='name of experiment')
    parser.add_argument('--prompt_file', type=str, help='file path of prompt')
    parser.add_argument('--model_name', type=str, help='name of model, gpt-35-turbo for example')
    args = parser.parse_args()

    with open(args.prompt_file) as f:
        prompt = "\n".join(f.readlines())

    exp_name = args.experiment_name
    n_samples = args.samples_per_benchmark
    model = args.model_name

    if args.area == "groundedness":
        start_mlflow_experiment(exp_name)
        ml_client = get_ml_client()
        exp_all_groundedness_benchmarks(ml_client, n_samples, model, prompt)
        end_mlflow_experiment()


if __name__ == '__main__':
    main()