from pathlib import Path
import pandas as pd
from prompt_builder import _PromptBuilder
from request import _request_prompt_batch
import json5
from typing import Dict, Generator, List, Union
from job import _Job

LABEL_KEYS = ["rating"]

class _JobManager:
    """Handles batching and job state management."""

    def __init__(
        self,
        output_fp: str,
        prompt_builder: _PromptBuilder
    ):
        self.output_fp = output_fp
        self.prompt_builder = prompt_builder
        self.job_idx = 0

    def submit_inputs(
        self,
        input_data_df: pd.DataFrame,
        request_args: dict,
        max_inputs: int,
        num_samples: int,
        min_score: int,
        max_score: int,
        **endpoint_args,
    ) -> pd.DataFrame:
        """Run inference over input dataset.

        Args:
            input_data_df: Spark DataFrame with input data.
            request_args (dict): Arguments to pass to API.
            max_inputs (int): Maximum number of inputs per prompt.
            num_samples (int): Number of samples to generate per prompt.
            ``**endpoint_args``: Arguments passed to _request_prompt_batch
        """
        # print(f"starting submit_inputs with endpoint_args={endpoint_args}")
        # Build batched prompts using inputs
        prompts = self.prompt_builder.generate_prompts(
            input_data_df,
            max_inputs=max_inputs,
            min_score=min_score,
            max_score=max_score
        )
        jobs = [
            self.make_job(prompt_data=prompt_data, request_params=request_args)
            for prompt_data in prompts
        ]

        # batch submission and parsing results
        # each job is one API call with 10 prompts concatenated together into a string separated by #TASK i
        # we parse output and write to file every JOB_BATCH jobs
        JOB_BATCH = 5
        start_idx = 0
        final_df = pd.DataFrame()
        while start_idx < len(jobs):
            end_idx = min(start_idx+JOB_BATCH, len(jobs))
            cur_batch = jobs[start_idx:end_idx]

            responses = _request_prompt_batch(
                jobs=cur_batch, **endpoint_args
            )

            # Parse responses for this batch of jobs
            parsed_dfs = self.parse_responses(responses, num_samples)

            batch_df = pd.DataFrame()
            #print(parsed_dfs)
            for parsed_df in parsed_dfs:
                batch_df = pd.concat([batch_df, parsed_df], axis=0)
            
            #display(batch_df)

            final_df = pd.concat([final_df, batch_df], axis=0)

            start_idx += JOB_BATCH
            if start_idx % 50 == 0:
                print(f"Finished {start_idx} out of {len(jobs)} jobs")
        
        return final_df
    
    @staticmethod
    def decode_example(
        example: str,
        label_keys: List[str],
    ) -> Dict[str, Union[int, bool, str]]:
        """Decode example from an encoding format.

        :param example: example to decode
        :param key_order: ordering of keys printed to string
        :return: decoded example
        """
        example = example.strip()
        start = example.rfind("{") # note this wouldn't work if returned result is a dictionary
        end = example.rfind("}")
        if start == -1:
            return {"malformed":"no left curly brace"}
        if end == -1:
            raise {"malformed":"no right curly brace"}
        
        example = json5.loads(example[start:end+1])

        # check if label keys are in example
        for label_key in label_keys:
            if label_key not in example:
                return {"malformed": "label key not in example"}
        return example

    def parse_responses(
        self,
        responses: List[_Job],
        num_samples: int,
    ) -> Generator[_Job, None, None]:
        """Given a stream of job responses, parse them."""
        for job in responses:
            label_keys = LABEL_KEYS
            job.response_data["label_keys"] = label_keys

            n_inputs = len(job.prompt_data.input_idx)

            if job.response_data["finish_reason"] == ["error"]:
                job.status = "endpoint error"
                print(
                    "Cannot parse job due to finish_reason=error: "
                    f"{job.response_data}"
                )
                continue

            assert len(job.response_data["samples"]) == num_samples, (
                f"Expected {num_samples} samples, "
                f"got {len(job.response_data['samples'])}"
            )

            output_examples = []
            num_failed = 0
            for sample in job.response_data["samples"]:
                # try to split the output into {n_inputs} examples
                try:
                    sample_examples = self.prompt_builder.split_output_examples(sample)

                    if len(sample_examples) < n_inputs:
                        raise ValueError(
                            f"Expected at least {n_inputs} examples, "
                            f"but got {len(sample_examples)}"
                        )

                except Exception:  # noqa: B902, F841
                    output_examples.append(None)
                    print(f"Failed splitting output into examples: {sample}")
                    num_failed += 1
                    continue

                # try to decode each example and check for the label keys
                try:
                    sample_examples_parsed = []
                    for example in sample_examples:
                        decoded = self.decode_example(example, label_keys)
                        if "malformed" not in decoded:
                            sample_examples_parsed.append(decoded)
                    output_examples.append(sample_examples_parsed)
                except Exception as _:  # noqa: B902, F841
                    output_examples.append([])
                    print(f"Failed decoding examples: {sample_examples}")
                    num_failed += 1

            if num_failed == num_samples:
                job.status = "parsing failed"
                yield pd.DataFrame()
            else:
                job.status = "success"
                job.response_data["output_examples"] = output_examples
                # skip if decoding error
                if output_examples[0] == None:
                    yield pd.DataFrame()
                    continue
                # construct a df
                df_input = pd.DataFrame(job.prompt_data.input_examples)
                df_output = pd.DataFrame({key: [res[key] for res in output_examples[0]] for key in LABEL_KEYS})
                if df_input.shape[0] != df_output.shape[0]:
                    print(f"Problematic: Output number {df_output.shape[0]} is different from input {df_input.shape[0]}")
                    print(job.response_data["samples"])
                    # discard this batch
                    yield pd.DataFrame()
                    continue
                df = pd.concat([df_input, df_output],axis=1)
                # first write to a file
                output_file_path = Path(self.output_fp)
                output_file_path.parent.mkdir(parents=True, exist_ok=True)
                df.to_csv(self.output_fp, mode='a', header=False)
                #print(f"written {df.shape[0]} records")
                yield df

    def make_job(self, **job_params) -> _Job:
        """Make job, tracking job_idx internally."""
        new_job = _Job(job_idx=self.job_idx, **job_params)
        self.job_idx += 1
        return new_job