import pandas as pd
import re
from collections import OrderedDict
import jinja2
from tokenizer import _Tokenizer
from typing import Dict, Generator, List, Optional, Tuple, Union
from prompt_data import _PromptData
import json5
from constants import CONTEXT, ANSWER

OUTPUT_SPLITTING_REGEX = r"[# ]*Task #*\d+:?"


class _InputSample:
    def __init__(self, index, encoded_json):
        self.index = index
        self.encoded_json = encoded_json

class _PromptBuilder:
    def __init__(
        self,
        labeling_guidelines: str,
        model_name: str,
        max_shots: int,
        max_tokens: int,
        min_input_examples: int = 1,
    ):
        """Batches prompts according to limits."""
        self.labeling_guidelines = jinja2.Template(
            self.format_batch_template(labeling_guidelines), undefined=jinja2.StrictUndefined
        )
        self.model_name = model_name
        self.max_tokens = max_tokens
        self.max_shots = max_shots
        self.min_input_examples = min_input_examples
        self.tokenizer = _Tokenizer(
            model_name=model_name,
        )

    def generate_prompts(
        self,
        input_data_df: pd.DataFrame,
        max_inputs: int,
        min_score: int,
        max_score: int
    ) -> Generator[_PromptData, None, None]:
        """Generate prompts from input_pool.

        Prioritizes fitting as many examples as possible into the prompt
        while staying below token limits.
        If an example can't fit into the prompt, then it is run on its own.

        Args:
            input_data_df (DataFrame): DataFrame of input examples
            max_inputs (int): maximum number of input examples to use

        Returns:
            prompt_data: list of PromptDatas containing the prompt,
                the index of input examples, and the number of tokens.
        """
        input_data = input_data_df.to_dict(orient="records")
        input_data_length = len(input_data)
        next_index = 0
        stop_index = min(max_inputs, input_data_length)

        while next_index < input_data_length:
            input_idx = list(range(next_index, stop_index))
            input_examples = input_data[next_index:stop_index]

            # Build prompt given input, shot, and token limits
            prompt, n_tokens, n_inputs = self.build_prompt_with_limits(
                input_examples,
                min_score,
                max_score
            )

            # send prompt
            next_index += n_inputs
            stop_index = min(next_index + max_inputs, input_data_length)
            input_idx = input_idx[:n_inputs]
            input_examples = input_examples[:n_inputs]
            yield _PromptData(
                input_idx=input_idx,
                input_examples=input_examples,
                prompt=prompt,
                n_tokens_estimate=n_tokens,
            )

    def build_prompt_with_limits(
        self,
        input_data: List[Dict[str, str]],
        min_score: int,
        max_score: int
    ) -> Tuple[str, int, int]:
        """Reduces batch examples until the prompt is within token limits.

        Reduces input_examples if maximum tokens are reached.

        Args:
            input_examples: list of input examples

        Returns:
            prompt: constructed prompt
            n_tokens: number of tokens within the prompt
            input_examples: number of input examples that can be fit
        """
        # build prompt with all examples
        prompt = self.build_prompt(input_data, min_score, max_score)
        n_tokens = self.tokenizer.count_tokens(prompt)

        # reduce input examples iteratively until minimum is hit
        while n_tokens > self.max_tokens and len(input_data) > self.min_input_examples:
            input_data = input_data[:-1]
            prompt = self.build_prompt(input_data)
            n_tokens = self.tokenizer.count_tokens(prompt)

        return prompt, n_tokens, len(input_data)

    @staticmethod
    def encode_example(
        example: Dict[str, Union[int, bool, str]],
        key_order: Optional[List[str]] = None,
        indent: Optional[int] = 2,
    ) -> str:
        """
        Encode examples into JSON.

        :param example: example to encode
        :param key_order: ordering of keys printed to string
        :param indent: number of spaces indented at each level
        :return: encoded example
        """
        if key_order:
            example = OrderedDict([(key, example[key]) for key in key_order])

        # Dump JSON with keys double-quoted and final comma removed
        return json5.dumps(
            example, indent=indent, quote_keys=True, trailing_commas=False
        )
    
    @staticmethod
    def format_batch_template(
        custom_prompt: str
    ) -> str:
        TASK_INJECTION_JINJA_TEMPLATE = """
        {% for sample in input_samples %}
        ## Task #{{ sample.index }}:
        {{ sample.encoded_json }}
        {% endfor %}
        """

        FINAL_ANNOTATION_REMINDER = "Reminder: The return values for each task should be correctly formatted dictionaries with the key \"rating\". Do not repeat the context and question. The value of \"rating\" should be an integer between {{min_score}} and {{max_score}}."  # noqa: E501

        prompt = "\n\n".join([
            custom_prompt,
            "## Actual Tasks",
            TASK_INJECTION_JINJA_TEMPLATE,
            FINAL_ANNOTATION_REMINDER,
            "## Actual Task Output:",
            ""
        ])
        return prompt

    def build_prompt(
        self,
        input_examples: List[Dict[str, str]],
        min_score,
        max_score
    ) -> str:
        """Build prompt from input_examples.

        Encode examples into JSON format.

        Args:
            input_examples: list of input examples

        Returns:
            prompt: constructed prompt
        """
        # create input prompt using pattern

        input_samples = [
            _InputSample(
                index=i,
                encoded_json=self.encode_example(input_example, [CONTEXT, ANSWER]),
            )
            for (i, input_example) in enumerate(input_examples, start=1)
        ]

        return self.labeling_guidelines.render(
            input_samples=input_samples,
            min_score=min_score,
            max_score=max_score
        )

    def split_output_examples(self, output_str: str) -> List[str]:
        """Attempt to split the output into a list of examples.

        Args:
            output_str: output examples

        Returns:
            output_examples: list of output examples
        """
        output_str = output_str.strip()
        output_examples = [
            ex.strip()
            for ex in re.split(OUTPUT_SPLITTING_REGEX, output_str)
            if ex.strip()
        ]
        
        return output_examples