import time
import requests
from job import _Job
from typing import List
from job import _HTTPClientWithRetry
from constants import MIN_REQUEST_COUNT, IGNORE_FAILED_REQUESTS_COUNT


def _request_api(
    session,
    endpoint_url: str,
    token: str,
    **request_params,
):
    headers = {
        "Content-Type": "application/json",
    }

    headers["api-key"] = token

    time_start = time.time()

    headers["timeout_ms"] = "90000"

    with session.post(
        endpoint_url,
        headers=headers,
        json=request_params,
        timeout=300,
    ) as response:
        time_taken = str(time.time() - time_start)
        #print(f"Received response from endpoint: {response.__dict__}")
        #print(f"Time taken to receive response: {time_taken}")
        if response.status_code == 200:
            response_data = response.json()

            time_taken = time.time() - time_start
            parsed_response = {
                "samples": [r["message"]["content"] for r in response_data["choices"]],
                "finish_reason": [r["finish_reason"] for r in response_data["choices"]],
            }

            return parsed_response, time_taken

        elif response.status_code == 400:  # content moderation filtered out, just skip this batch
            time_taken = time.time() - time_start
            return {}, time_taken

        else:
            raise Exception(
                "Received unexpected HTTP status: "
                f"{response.status_code} {response.text}"
            )


def _query_inference_for_line(
    jobs: List[_Job],
    session: requests.Session,
    endpoint_url: str,
    token: str,
    request_error_rate_threshold: float,
    api_call_delay_sec: float
):
    # if we count too many errors, we stop and raise an exception
    error_count = 0
    request_count = 0
    outputs = []
    for job in jobs:
        request_count += 1

        # Copy request_data to avoid modifying the original dict.
        request_data = {
            "model": job.request_params["model"],
            "temperature": job.request_params["temperature"],
            "top_p": job.request_params["top_p"],
            "n": job.request_params["num_samples"],
            "max_tokens": job.request_params["max_tokens"],
            "frequency_penalty": job.request_params["frequency_penalty"],
            "presence_penalty": job.request_params["presence_penalty"],
            "messages": [{"role": "user", "content": job.prompt_data.prompt}],
        }

        #print("prompt to send in")
        #print(job.prompt_data.prompt)

        if job.request_params.get("stop", None):
            request_data["stop"] = job.request_params["stop"]

        response = {}
        try:
            response, time_taken = _request_api(
                session=session,
                endpoint_url=endpoint_url,
                token=token,
                **request_data,
            )

            # Append time taken to the line
            response["response_time_sec"] = time_taken

        except Exception as e:  # noqa: B902
            response["finish_reason"] = ["error"]
            response["error"] = [str(e)]

            error_count += 1
            error_rate = error_count / request_count

            # if we count too many errors, we stop and raise an exception
            if error_count > IGNORE_FAILED_REQUESTS_COUNT or (
                request_count >= MIN_REQUEST_COUNT
                and error_rate >= request_error_rate_threshold
            ):
                raise e

        # skip job if no response due to content moderation
        if "samples" in response:
            job.response_data = response
            outputs.append(job)

        # Sleep between consecutive requests to avoid rate limit
        time.sleep(api_call_delay_sec)
    return outputs


def _request_prompt_batch(
    jobs: List[_Job],
    token: str,
    model: str,
    request_error_rate_threshold: float = 0.5,
    api_call_delay_sec: float = 0.5,
    api_call_retry_backoff_factor: int = 3,
    api_call_retry_max_count: int = 3,
    api_call_max_parallel_count: int = 1
) -> List[_Job]:

    endpoint_url = f"https://aoai-raidev.openai.azure.com/openai/deployments/{model}/chat/completions?api-version=2023-03-15-preview"

    #print(f"Determined endpoint URL {endpoint_url}")

    httpClient = _HTTPClientWithRetry(
        n_retry=api_call_retry_max_count,
        backoff_factor=api_call_retry_backoff_factor,
    )

    with httpClient.client as session:
        if len(jobs) == 0:
            return []  # queue is empty

        return _query_inference_for_line(
            jobs,
            session,
            endpoint_url,
            token,
            request_error_rate_threshold,
            api_call_delay_sec
        )